class YooMoneyError(Exception):
    """Basic class"""


class IllegalParamError(YooMoneyError):
    message = "Invalid parameter value {}"


class IllegalParamType(IllegalParamError):
    def __init__(self, ):
        super().__init__(self.message.format("type"))


class IllegalParamStartRecord(IllegalParamError):
    def __init__(self, ):
        super().__init__(self.message.format("start_record"))


class IllegalParamRecords(IllegalParamError):
    def __init__(self, ):
        super().__init__(self.message.format("records"))


class IllegalParamLabel(IllegalParamError):
    def __init__(self, ):
        super().__init__(self.message.format("label"))


class IllegalParamFromDate(IllegalParamError):
    def __init__(self, ):
        super().__init__(self.message.format("from_date"))


class IllegalParamTillDate(IllegalParamError):
    def __init__(self, ):
        super().__init__(self.message.format("till_date"))


class IllegalParamOperationId(IllegalParamError):
    def __init__(self, ):
        super().__init__(self.message.format("operation_id"))


class IllegalParamDetails(IllegalParamError):
    def __init__(self):
        super().__init__(self.message.format('details'))


class TechnicalError(YooMoneyError):
    message = "Technical error, try calling the operation again later. Error code - {error_code}"

    def __init__(self, error_code: str):
        super().__init__(self.message.format(error_code=error_code))


class InvalidRequest(YooMoneyError):
    message = "Required query parameters are missing or have incorrect or invalid values"

    def __init__(self, ):
        super().__init__(self.message)


class UnauthorizedClient(YooMoneyError):
    message = "Invalid parameter value 'client_id' or 'client_secret', or the application" \
              " does not have the right to request authorization (for example, YooMoney blocked it 'client_id')"

    def __init__(self, ):
        super().__init__(self.message)


class InvalidGrant(YooMoneyError):
    message = "In issue 'access_token' denied. YuMoney did not issue a temporary token, " \
              "the token is expired, or this temporary token has already been issued " \
              "'access_token' (repeated request for an authorization token with the same temporary token)"

    def __init__(self, ):
        super().__init__(self.message)


class EmptyToken(YooMoneyError):
    message = "Response token is empty. Repeated request for an authorization token"

    def __init__(self, ):
        super().__init__(self.message)


class InvalidToken(YooMoneyError):
    message = "Token is not valid, or does not have the appropriate rights"

    def __init__(self, ):
        super().__init__(self.message)



from dataclasses import dataclass, field

from .balance_details import BalanceDetails
from .card import Card
from ..enums.account import *


@dataclass
class Account:
    id: str = field(default="")
    balance: float = field(default=0.0)
    account_type: AccountType | None = field(default=None)
    account_status: AccountStatus | None = field(default=None)
    balance_details: BalanceDetails | None = field(default=None)
    cards_linked: list[Card] = field(default_factory=list)
    currency: int = field(default=643)
    identified: bool = field(default=False)



from os.path import join

from .context_session import ContextSession
from ..exceptions import InvalidToken
from ..globals import API_URL


class ContextMethod(ContextSession):
    def __init__(self, token: str, method: str, http_method: str = "POST", **kwargs):
        self.token = token
        self.headers = {
            'Authorization': 'Bearer ' + self.token,
            'Content-Type': 'application/x-www-form-urlencoded'
        }

        super().__init__(
            url=join(API_URL, method),
            method=http_method,
            headers=self.headers,
            **kwargs
        )

    async def __aenter__(self) -> dict:
        response = await super().__aenter__()
        response_json = await response.json()

        if not response_json:
            raise InvalidToken()

        return response_json



from datetime import datetime

from .exceptions import TechnicalError
from .globals import ERROR_CODES


def raise_error(error_code: str) -> None:
    if error_code in ERROR_CODES:
        raise ERROR_CODES[error_code]()
    else:
        raise TechnicalError(error_code)


def convert_datetime_to_str(date: datetime) -> str | None:
    if date is not None:
        return "{Y}-{m}-{d}T{H}:{M}:{S}".format(
            Y=str(date.year),
            m=str(date.month),
            d=str(date.day),
            H=str(date.hour),
            M=str(date.minute),
            S=str(date.second)
        )

    return None



from dataclasses import field, fields
from datetime import datetime as dt

from aioyoomoney.data_classes.digital import *
from aioyoomoney.enums.operation import *


@dataclass
class Operation:
    operation_id: str | None = field(default=None)
    status: OperationStatus | None = field(default=None)
    datetime: dt | None = field(default=None)
    title: str | None = field(default=None)
    pattern_id: str | None = field(default=None)
    direction: OperationDirection | None = field(default=None)
    amount: float | None = field(default=None)
    label: str | None = field(default=None)
    type: OperationType | None = field(default=None)
    kwargs: dict = field(default_factory=dict)

    @classmethod
    def serialize_from_dict(cls, data: dict) -> "Operation":
        operation = cls()

        valid_keys = [_field.name for _field in fields(operation)]  # не нравится
        for key, value in data.items():
            if key in valid_keys:
                operation.__setattr__(key, value)
            else:
                operation.kwargs[key] = value

        return operation

    def __getitem__(self, item: str):
        if item in self.kwargs.keys():
            return self.kwargs.get(item)

        return self.__getattribute__(item)


@dataclass
class OperationDetails(Operation):
    details: str | None = field(default=None)
    codepro: bool | None = field(default=None)
    comment: str | None = field(default=None)
    message: str | None = field(default=None)


@dataclass
class OperationDetailsIncomingTransfer(OperationDetails):
    sender: str | None = field(default=None)

    def __post_init__(self):
        self.type = OperationType.INCOMING_TRANSFER


@dataclass
class OperationDetailsOutgoingTransfer(OperationDetails):
    recipient: str | None = field(default=None)
    recipient_type: RecipientType | None = field(default=None)
    amount_due: int | None = field(default=None)
    fee: int | None = field(default=None)

    def __post_init__(self):
        self.type = OperationType.OUTGOING_TRANSFER


@dataclass
class OperationDetailsPaymentShop(OperationDetailsOutgoingTransfer):
    digital_goods: DigitalGoods | None = field(default=None)

    def __post_init__(self):
        self.type = OperationType.PAYMENT_SHOP



from .account import Account
from .balance_details import BalanceDetails
from .card import Card
from .digital import DigitalBonus, DigitalArticle, DigitalGoods
from .history import History
from .operation import Operation
from .operation import OperationDetails, Operation, OperationDetailsOutgoingTransfer



from .exceptions import *

API_URL = "https://yoomoney.ru/api/"
ERROR_CODES = {
    "illegal_param_type": IllegalParamType,
    "illegal_param_start_record": IllegalParamStartRecord,
    "illegal_param_records": IllegalParamRecords,
    "illegal_param_label": IllegalParamLabel,
    "illegal_param_from": IllegalParamFromDate,
    "illegal_param_till": IllegalParamTillDate,
    "illegal_param_details": IllegalParamDetails,
    "illegal_param_operation_id": IllegalParamOperationId,
    
    "invalid_request": InvalidRequest,
    "unauthorized_client": UnauthorizedClient,
    "invalid_grant": InvalidGrant
}



from dataclasses import dataclass, field


@dataclass
class BalanceDetails:
    total: float | None = field(default=None)
    available: float | None = field(default=None)
    deposition_pending: float | None = field(default=None)
    blocked: float | None = field(default=None)
    debt: float | None = field(default=None)
    hold: float | None = field(default=None)



from enum import StrEnum


class OperationStatus(StrEnum):
    SUCCESS = "success"
    REFUSED = "refused"
    IN_PROGRESS = "in_progress"


class OperationDirection(StrEnum):
    IN = "in"
    OUT = "out"


class OperationType(StrEnum):
    PAYMENT_SHOP = "payment-shop"
    OUTGOING_TRANSFER = "outgoing-transfer"
    DEPOSITION = "deposition"
    INCOMING_TRANSFER = "incoming-transfer"


class RecipientType(StrEnum):
    ACCOUNT = "account"
    PHONE = "phone"
    EMAIL = "email"



from dataclasses import dataclass, field

from .operation import Operation


@dataclass
class History:
    next_record: int | None = field(default=None)
    operations: list[Operation] = field(default_factory=list)



from datetime import datetime

from aioyoomoney.methods import *
from .data_classes import *
from .enums import *
from .methods.operation_details import OPERATIONS_TYPE


class Client:
    def __init__(self, token: str):
        self.token = token

    async def account_info(self) -> Account:
        async with AccountMethod(self.token) as account:
            return account

    async def operation_history(
            self,
            type: OperationHistoryType = None,
            label: str = None,
            from_date: datetime = None,
            till_date: datetime = None,
            start_record: str = None,
            records: int = None,
            details: bool = False
    ) -> History:
        async with HistoryMethod(
                token=self.token,
                type=type,
                label=label,
                from_date=from_date,
                till_date=till_date,
                start_record=start_record,
                records=records,
                details=details,
        ) as history:
            return history

    async def operation_details(self, operation_id: str) -> OPERATIONS_TYPE:
        async with OperationDetailsMethod(
                token=self.token,
                operation_id=operation_id,
        ) as operation_details:
            return operation_details




import requests

from urllib.parse import quote_plus
from yarl import URL

from .exceptions import EmptyToken
from .utils import raise_error


def authorize(client_id: str, redirect_uri: str, client_secret: str, scope: list[str]) -> str:
    redirect_uri = quote_plus(redirect_uri)
    url = (
        "https://yoomoney.ru/oauth/authorize?client_id={client_id}&response_type=code"
        "&redirect_uri={redirect_uri}&scope={scope}"
        .format(client_id=client_id, redirect_uri=redirect_uri, scope='%20'.join([str(elem) for elem in scope]))
    )

    headers = {
        'Content-Type': 'application/x-www-form-urlencoded'
    }

    response = requests.post(url, headers=headers)
    if response.status_code != 200:
        raise

    print("Visit this website and confirm the application authorization request:")
    print(response.url)

    code_or_url = URL(input("Enter redirected url (https://yourredirect_uri?code=XXXXXXXXXXXXX) or just code: "))
    code = code_or_url.query.get("code")
    if not code:
        code = code_or_url

    url = (
        "https://yoomoney.ru/oauth/token?code={code}&client_id={client_id}&"
        "grant_type=authorization_code&redirect_uri={redirect_uri}&client_secret={client_secret}"
        .format(code=code, client_id=client_id, redirect_uri=redirect_uri, client_secret=client_secret)
    )

    data = requests.post(url, headers=headers).json()
    if "error" in data:
        raise_error(data["error"])

    if data['access_token'] == "":
        raise EmptyToken()

    return data['access_token']



from datetime import datetime
from pprint import pprint

from .operation_details import OperationDetailsMethod
from ..base.context_method import ContextMethod
from ..data_classes import *
from ..enums.history import OperationHistoryType
from ..exceptions import (
    IllegalParamFromDate,
    IllegalParamTillDate,
)
from ..utils import raise_error, convert_datetime_to_str


class HistoryMethod(ContextMethod):
    def __init__(
            self, token: str = None,
            type: OperationHistoryType = None,
            label: str = None,
            from_date: datetime = None,
            till_date: datetime = None,
            start_record: str = None,
            records: int = None,
            details: bool = False
    ):
        self.details = details

        try:
            from_date = convert_datetime_to_str(from_date)
        except:
            raise IllegalParamFromDate()

        try:
            till_date = convert_datetime_to_str(till_date)
        except:
            raise IllegalParamTillDate()

        payload = {
            "type": type,
            "label": label,
            "from": from_date,
            "till": till_date,
            "start_record": start_record,
            "records": records,
            "details": str(details).lower()
        }

        payload = {k: v for k, v in payload.items() if v is not None}
        super().__init__(token, "operation-history", data=payload)

    async def __aenter__(self) -> History:
        data = await super().__aenter__()

        if "error" in data:
            raise_error(data["error"])

        history = History()

        if "next_record" in data:
            history.next_record = int(data["next_record"])

        for operation_data in data["operations"]:
            operation = OperationDetailsMethod.get_operation_from_dict(operation_data, self.details)

            history.operations.append(operation)

        return history



from enum import StrEnum


class AccountStatus(StrEnum):
    ANONYMOUS = "anonymous"
    NAMED = "named"
    IDENTIFIED = "identified"


class AccountType(StrEnum):
    PROFESSIONAL = "professional"
    PERSONAL = "personal"



from aiohttp import ClientSession, ClientResponse
from yarl import URL


class ContextSession:
    def __init__(self, url: str | URL, method: str, **kwargs):
        self._session = None
        self._response = None

        self.__url = url
        self.__method = method
        self.__kwargs = kwargs

    async def __aenter__(self) -> ClientResponse:
        self._session = ClientSession()
        self._response = await self._session.request(
            self.__method,
            self.__url,
            **self.__kwargs
        )

        return self._response

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self._response.close()
        await self._session.close()



from dataclasses import dataclass


@dataclass
class DigitalArticle:
    merchant_article_id: str
    serial: str
    secret: str


@dataclass
class DigitalBonus:
    serial: str
    secret: str


@dataclass
class DigitalGoods:
    articles: list[DigitalArticle]
    bonuses: list[DigitalBonus]



from ..base.context_method import ContextMethod
from ..data_classes import *


class AccountMethod(ContextMethod):
    def __init__(self, token: str):
        super().__init__(token, "account-info")

    async def __aenter__(self) -> Account:
        data = await super().__aenter__()

        return self._get_account_object(data)

    def _get_account_object(self, data: dict) -> Account:
        account = Account()
        account.id = data['account']
        account.balance = data['balance']
        account.currency = data['currency']
        account.account_status = data['account_status']
        account.account_type = data['account_type']
        account.identified = data["identified"]

        if 'balance_details' in data:
            balance_details = data['balance_details']
            account.balance_details = BalanceDetails()
            account.balance_details.available = float(balance_details.get('available', 0.0))
            account.balance_details.blocked = float(balance_details.get('blocked', 0.0))
            account.balance_details.debt = float(balance_details.get('debt', 0.0))
            account.balance_details.deposition_pending = float(balance_details.get('deposition_pending', 0.0))
            account.balance_details.total = float(balance_details.get('total', 0.0))
            account.balance_details.hold = float(balance_details.get('hold', 0.0))

        if 'cards_linked' in data:
            for card_linked in data['cards_linked']:
                card = Card(
                    pan_fragment=card_linked['pan_fragment'],
                    type=card_linked['type']
                )
                account.cards_linked.append(card)

        return account



from dateutil.parser import parse as dateutil_parse

from aioyoomoney.base.context_method import ContextMethod
from aioyoomoney.data_classes.operation import *
from aioyoomoney.enums.operation import OperationType
from aioyoomoney.utils import raise_error


OPERATIONS_TYPE = Operation | OperationDetails | OperationDetailsIncomingTransfer | OperationDetailsOutgoingTransfer | OperationDetailsPaymentShop


class OperationDetailsMethod(ContextMethod):
    def __init__(self, token: str, operation_id: str):
        super().__init__(
            token,
            "operation-details",
            "POST",
            data={
                "operation_id": operation_id
            }
        )

    async def __aenter__(self) -> OPERATIONS_TYPE:
        data = await super().__aenter__()

        if "error" in data:
            raise_error(data["error"])

        return self.get_operation_from_dict(data)

    @staticmethod
    def get_operation_from_dict(data: dict, details=True) -> OPERATIONS_TYPE:
        datetime = dateutil_parse(data.get("datetime", ""))

        data["datetime"] = datetime

        if not details:
            return Operation.serialize_from_dict(data)

        match data["type"]:
            case OperationType.PAYMENT_SHOP:
                if "digital_goods" in data:
                    articles = []
                    for product in data["digital_goods"]["article"]:
                        digital_article = DigitalArticle(merchant_article_id=product["merchantArticleId"],
                                                         serial=product["serial"],
                                                         secret=product["secret"])
                        articles.append(digital_article)

                    bonuses = []
                    for bonus in data["digital_goods"]["bonus"]:
                        digital_bonus = DigitalBonus(serial=bonus["serial"],
                                                     secret=bonus["secret"])
                        bonuses.append(digital_bonus)

                    data["digital_goods"] = DigitalGoods(articles=articles,
                                                         bonuses=bonuses)
                operation = OperationDetailsPaymentShop.serialize_from_dict(data)
            case OperationType.OUTGOING_TRANSFER:
                operation = OperationDetailsOutgoingTransfer.serialize_from_dict(data)
            case OperationType.INCOMING_TRANSFER:
                operation = OperationDetailsIncomingTransfer.serialize_from_dict(data)
            case _:
                operation = OperationDetails.serialize_from_dict(data)

        return operation



import aiohttp

from .base import ContextSession
from .enums.quickpay import PaymentType


QUICKPAY_URL = "https://yoomoney.ru/quickpay/confirm"


# https://yoomoney.ru/docs/payment-buttons/using-api/forms
class Quickpay(ContextSession):
    def __init__(
            self,
            receiver: str,
            sum: float,
            payment_type: PaymentType = None,
            label: str = None,
            success_url: str = None,
    ):
        self.data = {
            "receiver": receiver,
            "quickpay-form": "button",
            "paymentType": payment_type,
            "sum": sum,
            "label": label,
            "successURL": success_url,
        }

        self.data = {k: v for k, v in self.data.items() if v is not None}

        self.url = None

        super().__init__(QUICKPAY_URL, "POST", data=self.data)

    async def __aenter__(self) -> "Quickpay":
        response = await super().__aenter__()
        self.url = response.url

        return self
